package com.qianwankj.vt.vonage.flutter_vonage

import android.annotation.SuppressLint
import android.content.Context
import android.media.AudioManager
import android.net.Uri
import android.os.Handler
import android.os.Looper
import com.nexmo.client.*
import com.nexmo.client.request_listener.NexmoApiError
import com.nexmo.client.request_listener.NexmoRequestListener
import io.flutter.plugin.common.EventChannel
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import java.util.*

/**
 * 返回给flutter层信息
 */
class ResultEventInfo(private val type: String, private val info: String? = null) {
    override fun toString(): String {
        return "{\"type\":\"$type\",\"info\":\"${Uri.encode(info)}\"}"
    }
}

class CallUser(private val mid: String?, private val uid: String?, private val name: String?) {
    override fun toString(): String {
        return "{\"mid\":\"$mid\",\"uid\":\"$uid\",\"name\":\"$name\"}"
    }
}

/**
 * 1. 初始化
 * 2. 登录
 * 3. 拨打电话
 * 4. 发送信息
 */
class VonageUtils(private val context: Context) {

    private val mHandler = Handler(Looper.getMainLooper())
    private var mNexmoCall: NexmoCall? = null
    private var mNexmoConversation: NexmoConversation? = null
    var eventSink: EventChannel.EventSink? = null


    private val mNexmoCallEventListener = object : NexmoCallEventListener {
        override fun onMemberStatusUpdated(
            callStatus: NexmoCallMemberStatus?,
            member: NexmoMember?
        ) {
            VtLogUtil.d("NexmoCallMemberStatus: $callStatus")

            if (callStatus == NexmoCallMemberStatus.COMPLETED || callStatus == NexmoCallMemberStatus.CANCELLED) {
                // 已挂断
                val callId = mNexmoCall?.conversation?.conversationId ?: ""
                val info = "{\"callId\":\"$callId\"}"

                removeCurNexmoCall()
                callback(ResultEventInfo(Constants.CLOSE_CALL, info))
            }
        }

        override fun onMuteChanged(newState: NexmoMediaActionState?, member: NexmoMember?) {
        }

        override fun onEarmuffChanged(newState: NexmoMediaActionState?, member: NexmoMember?) {
        }

        override fun onDTMF(dtmf: String?, member: NexmoMember?) {
        }
    }

    private val messageListener = object : NexmoMessageEventListener {
        override fun onTextEvent(textEvent: NexmoTextEvent) {
            VtLogUtil.d("textEvent->$textEvent")
            handleIncomingEvent(textEvent)
        }

        override fun onAttachmentEvent(attachmentEvent: NexmoAttachmentEvent) {
            VtLogUtil.d("onAttachmentEvent->$attachmentEvent")
        }

        override fun onEventDeleted(deletedEvent: NexmoDeletedEvent) {
            VtLogUtil.d("deletedEvent->$deletedEvent")
        }

        override fun onSeenReceipt(seenEvent: NexmoSeenEvent) {
            VtLogUtil.d("seenEvent->$seenEvent")
        }

        override fun onDeliveredReceipt(deliveredEvent: NexmoDeliveredEvent) {
            VtLogUtil.d("deliveredEvent->$deliveredEvent")
        }

        override fun onTypingEvent(typingEvent: NexmoTypingEvent) {
            VtLogUtil.d("typingEvent->$typingEvent")
        }
    }

    private val customEventListener = NexmoCustomEventListener {
        VtLogUtil.d("customEvent->$it")
        handleIncomingEvent(it)
    }

    @SuppressLint("MissingPermission")
    fun onMethodCall(ctx: Context, call: MethodCall, result: MethodChannel.Result) {
        VtLogUtil.d("onMethodCall: ${call.method}")

        if (Constants.INIT == call.method) {
            VtLogUtil.open = call.argument<Boolean>("openLog") ?: true
            // 初始化
            NexmoClient.Builder().build(ctx)

            // 默认设置连接监听
            NexmoClient.get().setConnectionListener { status, reason ->
                run {
                    val info = "{\"status\":\"$status\",\"reason\":\"$reason\"}"
                    VtLogUtil.d("ConnectionListener: $info")

                    callback(ResultEventInfo(Constants.CONNECTION_RESULT, info))
                }
            }

            // 默认添加来电监听
            NexmoClient.get().addIncomingCallListener {
                VtLogUtil.d("addIncomingCallListener: 接收到新的来电了->$it")
                handleIncomingCall(it)
            }

            result.success(Constants.INIT)
            return
        }

        if (Constants.LOGIN == call.method) {
            val phone = call.argument<String>("phone") ?: ""
            val jwtToken = call.argument<String>("jwtToken") ?: ""

            NexmoClient.get().login(jwtToken)

            result.success(Constants.LOGIN)
            return
        }

        if (Constants.LOGOUT == call.method) {
            NexmoClient.get().logout()

            result.success(Constants.LOGOUT)
            return
        }

        if (Constants.CLOSE == call.method) {
            // TODO 释放资源

            result.success(Constants.CLOSE)
            return
        }

        if (Constants.UPDATE_PUSH_TOKEN == call.method) {
            val pushToken = call.argument<String>("pushToken") ?: ""
            // 更新推送token
            NexmoClient.get()
                .enablePushNotifications(pushToken, object : NexmoRequestListener<Void> {
                    override fun onSuccess(v: Void?) {
                        VtLogUtil.d("enablePushNotifications: 设置pushToken成功！")
                    }

                    override fun onError(apiError: NexmoApiError) {
                        VtLogUtil.e("enablePushNotifications: ${apiError.message}")
                    }
                })

            result.success(Constants.UPDATE_PUSH_TOKEN)
            return
        }

        if (Constants.HANDLE_PUSH == call.method) {
            // 处理推送数据
            val data = call.arguments as MutableMap<String, String>
            if (NexmoClient.isNexmoPushNotification(data)) {
                NexmoClient.get().processNexmoPush(data, object : NexmoPushEventListener {
                    override fun onIncomingCall(call: NexmoCall?) {
                        // 推送来电
                        VtLogUtil.d("FirebaseMessage:onIncomingCall() with: $call")
                        handleIncomingCall(call)
                    }

                    override fun onNewEvent(event: NexmoEvent?) {
                        // 推送事件
                        VtLogUtil.d("FirebaseMessage:onNewEvent() with: $event")
                        handleIncomingEvent(event);
                    }

                    override fun onError(apiError: NexmoApiError?) {
                        VtLogUtil.d("FirebaseMessage:onError() with: $apiError")
                    }
                })
            }
            result.success(Constants.HANDLE_PUSH)
            return
        }

        if (Constants.START_CHAT == call.method) {
            // 准备开始app内部聊天
            removeCurNexmoConversation()
            val conversationId = call.argument<String>("conversationId") ?: ""
            NexmoClient.get()
                .getConversation(conversationId, object : NexmoRequestListener<NexmoConversation> {
                    override fun onError(error: NexmoApiError) {
                        callback(ResultEventInfo(Constants.START_CHAT_RESULT, "-1"))
                    }

                    override fun onSuccess(result: NexmoConversation?) {
                        if (null == result) {
                            callback(ResultEventInfo(Constants.START_CHAT_RESULT, "-1"))
                            return
                        }

                        // 获取会话成功
                        mNexmoConversation = result
                        // 监听会话消息
                        mNexmoConversation?.addMessageEventListener(messageListener)
                        mNexmoConversation?.addCustomEventListener(customEventListener)

                        callback(ResultEventInfo(Constants.START_CHAT_RESULT, "1"))
                    }
                })
        }

        if (Constants.SEND_MESSAGE == call.method) {
            // 发送消息
            val type = call.argument<Int>("type") ?: 0 // 0 文本 1 图片 2地图
            val customData = call.arguments as HashMap<String, Any>

            val listener = object : NexmoRequestListener<Void?> {
                override fun onSuccess(v: Void?) {
                    VtLogUtil.d("Message sent success")
                    callback(ResultEventInfo(Constants.SEND_MSG_RESULT, "1"))
                }

                override fun onError(apiError: NexmoApiError) {
                    VtLogUtil.e("Error:Message not sent ${apiError.message}")
                    callback(ResultEventInfo(Constants.SEND_MSG_RESULT, "-1"))
                }
            }

            when (type) {
                0 -> {
                    mNexmoConversation?.sendCustomEvent("vt_txt", customData, listener)
                }
                1 -> {
                    mNexmoConversation?.sendCustomEvent("vt_image", customData, listener)
                }
                2 -> {
                    mNexmoConversation?.sendCustomEvent("vt_map", customData, listener)
                }
                else -> {
                    // 消息发送失败
                    callback(ResultEventInfo(Constants.SEND_MSG_RESULT, "-1"))
                }
            }

            result.success(Constants.SEND_MESSAGE)
            return
        }

        if (Constants.CALL_ANSWER == call.method) {
            // 接通
            mNexmoCall?.answer(object : NexmoRequestListener<NexmoCall> {
                override fun onSuccess(call: NexmoCall?) {
                    val callId = mNexmoCall?.conversation?.conversationId ?: ""
                    val info = "{\"type\":1,\"callId\":\"$callId\"}"

                    callback(ResultEventInfo(Constants.INCOMING_CALL_ANSWER, info))
                }

                override fun onError(apiError: NexmoApiError) {
                    val callId = mNexmoCall?.conversation?.conversationId ?: ""
                    val info = "{\"type\":-1,\"callId\":\"$callId\"}"

                    callback(ResultEventInfo(Constants.INCOMING_CALL_ANSWER, info))
                }
            })

            result.success(Constants.CALL_ANSWER)
            return
        }

        if (Constants.CALL_REJECT == call.method) {
            // 拒接
            mNexmoCall?.hangup(object : NexmoRequestListener<NexmoCall> {
                override fun onSuccess(call: NexmoCall?) {
                    val callId = mNexmoCall?.conversation?.conversationId ?: ""
                    val info = "{\"type\":1,\"callId\":\"$callId\"}"

                    removeCurNexmoCall()
                    callback(ResultEventInfo(Constants.INCOMING_CALL_REJECT, info))
                }

                override fun onError(apiError: NexmoApiError) {
                    val callId = mNexmoCall?.conversation?.conversationId ?: ""
                    val info = "{\"type\":-1,\"callId\":\"$callId\"}"

                    removeCurNexmoCall()
                    callback(ResultEventInfo(Constants.INCOMING_CALL_REJECT, info))
                }
            })

            result.success(Constants.CALL_REJECT)
            return
        }

        if (Constants.CALL == call.method) {
            // 拨号
            val type = call.argument<String>("type") ?: ""
            val from = call.argument<String>("from") ?: ""
            val to = call.argument<String>("to") ?: ""

            removeCurNexmoCall()

            val listener = object : NexmoRequestListener<NexmoCall> {
                override fun onError(error: NexmoApiError) {
                    val info = "{\"type\":-1}"
                    callback(ResultEventInfo(Constants.OUTGOING_CALL, info))
                }

                override fun onSuccess(result: NexmoCall?) {
                    if (null == result) {
                        val info = "{\"type\":-1}"
                        callback(ResultEventInfo(Constants.OUTGOING_CALL, info))
                        return
                    }

                    mNexmoCall = result
                    // 设置监听
                    mNexmoCall?.addCallEventListener(mNexmoCallEventListener)

                    val callId = result.conversation.conversationId

                    // 待接听
                    val info =
                        "{\"type\":1,\"callId\":\"$callId\",\"from\":\"$from\",\"to\":\"$to\"}"
                    callback(ResultEventInfo(Constants.OUTGOING_CALL, info))
                }
            }

            when (type) {
                "IN_APP" -> {
                    NexmoClient.get().call(to, NexmoCallHandler.IN_APP, listener)
                }
                "SERVER" -> {
                    NexmoClient.get().call(to, NexmoCallHandler.SERVER, listener)
                }
                else -> {
                    // 不支持的类型
                    val info = "{\"type\":-1}"
                    callback(ResultEventInfo(Constants.OUTGOING_CALL, info))
                }
            }

            result.success(Constants.CALL)
            return
        }

        if (Constants.CALL_END == call.method) {
            // 挂断
            mNexmoCall?.hangup(object : NexmoRequestListener<NexmoCall> {
                override fun onSuccess(call: NexmoCall?) {
                    val callId = mNexmoCall?.conversation?.conversationId ?: ""
                    val info = "{\"type\":1,\"callId\":\"$callId\"}"

                    callback(ResultEventInfo(Constants.OUTGOING_CALL_HANGUP, info))
                }

                override fun onError(apiError: NexmoApiError) {
                    val callId = mNexmoCall?.conversation?.conversationId ?: ""
                    val info = "{\"type\":-1,\"callId\":\"$callId\"}"

                    callback(ResultEventInfo(Constants.OUTGOING_CALL_HANGUP, info))
                }
            })

            result.success(Constants.CALL_END)
            return
        }

        if (Constants.CALL_EVENT_MUTE == call.method) {
            // 是否静音
            val mute = call.argument<Boolean>("mute") ?: false
            mNexmoCall?.mute(mute)

            result.success(Constants.CALL_EVENT_MUTE)
            return
        }

        if (Constants.CALL_EVENT_SPEAK == call.method) {
            // 是否外放
            val speak = call.argument<Boolean>("speak") ?: false
            openSpeaker(speak)
            result.success(Constants.CALL_EVENT_SPEAK)
            return
        }

        if (Constants.CALL_EVENT_RECORD == call.method) {
            // TODO 是否录音
            val record = call.argument<Boolean>("record") ?: false

            result.success(Constants.CALL_EVENT_RECORD)
            return
        }

        if (Constants.CALL_EVENT_DTMF == call.method) {
            // 发送指令
            val dtmf = call.argument<String>("dtmf") ?: ""
            mNexmoCall?.sendDTMF(dtmf, object : NexmoRequestListener<Void> {
                override fun onError(error: NexmoApiError) {
                }

                override fun onSuccess(result: Void?) {
                }

            })

            result.success(Constants.CALL_EVENT_DTMF)
            return
        }

        result.notImplemented()
    }

    /**
     * 移除当前会话
     */
    private fun removeCurNexmoConversation() {
        try {
            mNexmoConversation?.removeMessageEventListener(messageListener)
            mNexmoConversation?.removeCustomEventListener(customEventListener)

        } catch (e: Exception) {
            VtLogUtil.e("removeCurNexmoConversation: ${e.message}")
        }

        mNexmoConversation = null
    }

    // 移除现有状态
    private fun removeCurNexmoCall() {
        try {
            mNexmoCall?.removeCallEventListener(mNexmoCallEventListener)

        } catch (e: Exception) {
            VtLogUtil.e("removeCurNexmoCall: ${e.message}")
        }

        mNexmoCall = null
    }

    /**
     * 处理来电
     */
    private fun handleIncomingCall(nexmoCall: NexmoCall?) {
        if (null == nexmoCall) {
            VtLogUtil.e("handleIncomingCall: nexmoCall is null!")
            return
        }

        removeCurNexmoCall()

        mNexmoCall = nexmoCall
        mNexmoCall?.addCallEventListener(mNexmoCallEventListener)

        // 来电待接通
        val cur = nexmoCall.myMember // 当前用户
        val inUser = nexmoCall.allMembers.find { it.memberId != cur.memberId } // 来电的用户

        val callId = nexmoCall.conversation.conversationId;
        val sender = CallUser(inUser?.memberId, inUser?.user?.id, inUser?.user?.name)
        val receiver = CallUser(cur.memberId, cur.user.id, cur.user.name)
        val info = "{\"callId\":\"$callId\",\"sender\":\"$sender\",\"receiver\":\"$receiver\"}"

        callback(ResultEventInfo(Constants.INCOMING_CALL, info))
    }

    /**
     * 处理入站事件
     */
    private fun handleIncomingEvent(event: NexmoEvent?) {
        if (null == event) {
            VtLogUtil.e("handleIncomingEvent: event is null!")
            return
        }

        if (event is NexmoCustomEvent) {
            if (event.customType == "vt_txt" || event.customType == "vt_image" || event.customType == "vt_map") {
                // 是自定义的类型
                val type = event.data["type"]
                val from = event.data["from"]
                val to = event.data["to"]
                val content = event.data["content"]
                val address = event.data["address"]
                val lat = event.data["lat"]
                val lon = event.data["lon"]

                // 回传消息
                val info =
                    "{\"type\":\"$type\",\"from\":\"$from\",\"to\":\"$to\",\"content\":\"$content\",\"address\":\"$address\",\"lat\":$lat,\"lon\":$lon}"
                callback(ResultEventInfo(Constants.INCOMING_MSG, info))
            }
        }
    }

    var curVolume: Int = 0

    // 检查是否关闭扬声器了，有些机子开机时，默认是关闭状态，此时需要打开，否则出不了声
    private fun openSpeaker(open: Boolean) {
        try {
            val audioManager = context.getSystemService(Context.AUDIO_SERVICE) as AudioManager
            if (open) {
                // 打开扬声器
                if (!audioManager.isSpeakerphoneOn) {
                    // 记录当前的声音大小
                    curVolume = audioManager.getStreamVolume(AudioManager.STREAM_VOICE_CALL)

                    audioManager.mode = AudioManager.MODE_IN_CALL
                    audioManager.isSpeakerphoneOn = true
                    audioManager.setStreamVolume(
                        AudioManager.STREAM_VOICE_CALL,
                        audioManager.getStreamMaxVolume(AudioManager.STREAM_VOICE_CALL),
                        AudioManager.STREAM_VOICE_CALL
                    )
                }

            } else {
                // 关闭扬声器
                if (audioManager.isSpeakerphoneOn) {
                    audioManager.mode = AudioManager.MODE_IN_CALL
                    audioManager.isSpeakerphoneOn = false
                    audioManager.setStreamVolume(
                        AudioManager.STREAM_VOICE_CALL,
                        curVolume,
                        AudioManager.STREAM_VOICE_CALL
                    )
                }
            }
        } catch (e: Exception) {
            VtLogUtil.e(e.message)
        }
    }

    // 回传到flutter，需要在UI线程执行
    private fun callback(ev: ResultEventInfo) {
        mHandler.post {
            try {
                VtLogUtil.d("消息回调: $ev")

                eventSink?.success(ev.toString())
            } catch (e: Exception) {
                VtLogUtil.e("消息回调异常: ${e.message}")
            }
        }
    }

    fun destroy() {
        // TODO 销毁

    }

}